// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';

/// {@template authentication_base_data_source}
/// Base class for custom [AuthenticationRemoteDataSource] implementations.
/// It implements all methods as throwing [UnimplementedError]s.
/// {@endtemplate}
class AuthenticationBaseDataSource<Data>
    extends AuthenticationRemoteDataSource<Data> {
  /// {@macro authentication_base_data_source}
  const AuthenticationBaseDataSource();

  @override
  Future<void> confirmPasswordReset({
    required String code,
    required String newPassword,
  }) {
    throw UnimplementedError();
  }

  @override
  Future<void> delete() {
    throw UnimplementedError();
  }

  @override
  Future<Account> reauthenticate() {
    throw UnimplementedError();
  }

  @override
  Future<Account> refresh() {
    throw UnimplementedError();
  }

  @override
  Future<void> sendEmailVerification() {
    throw UnimplementedError();
  }

  @override
  Future<void> sendPasswordResetEmail({required String email}) {
    throw UnimplementedError();
  }

  @override
  Future<Account> signInAnonymously() {
    throw UnimplementedError();
  }

  @override
  Future<Account> signInWithEmailAndPassword({
    required String email,
    required String password,
  }) {
    throw UnimplementedError();
  }

  @override
  Future<Account> signInWithGoogle() {
    throw UnimplementedError();
  }

  @override
  Future<void> signOut() {
    throw UnimplementedError();
  }

  @override
  Future<Account> signUpWithEmailAndPassword({
    required String email,
    required String password,
  }) {
    throw UnimplementedError();
  }

  @override
  Future<Account> updateEmail({required String email}) {
    throw UnimplementedError();
  }

  @override
  Future<Account> updatePassword({required String password}) {
    throw UnimplementedError();
  }

  @override
  Future<bool> verifyPasswordResetCode({required String code}) {
    throw UnimplementedError();
  }
}
